/* ====================================================================
 * Copyright (c) 1997, 1998 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * mod_jserv.c: Implements part of Java Server API for Apache
 * by Alexei Kosut <akosut@apache.org>
 * parts based on mod_alias.c and mod_cgi.c
 *
 * For more information, see http://java.apache.org/
 */

#include "httpd.h"
#include "http_config.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"

/* Name of the class to run - not configurable */
#define JSERV_CLASS "org.apache.jserv.JServHandler"

/* Some defaults */
#define JSERV_DEFAULT_PORT 8007
#define JSERV_DEFAULT_SERVER inet_addr("127.0.0.1")
#define JSERV_DEFAULT_PROPERTIES "conf/servlets.properties"
#define JSERV_AUTH_CONF_SIZE 10

#define CP_SEP ":"

/* This is taken from mod_log_config */

static int xfer_flags = (O_WRONLY | O_APPEND | O_CREAT);
#if defined(__EMX__) || defined(WIN32)
/* OS/2 dosen't support users and groups */
static mode_t xfer_mode = (S_IREAD | S_IWRITE);
#else
static mode_t xfer_mode = (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
#endif


typedef struct {
    char *real;
    char *fake;
    int is_dir;
} jserv_entry;

typedef struct {
    array_header *dirs;
    array_header *envs;
    array_header *vhost;	/* Only set in the base host configuration
				 * contains pointer to the virtual host
				 * information to pass to java process.
				 */
    server_rec *host;		/* The server record for wich we are configured */
    char *classpath;

    char *binary;
    array_header *binary_args;	/* Extra argument on the commandline */

    char *properties;
    char *errlog;
    int port;
    unsigned long remote_server;

    char auth[JSERV_AUTH_CONF_SIZE];
    int auth_export;
    int manual;
} jserv_conf;

module jserv_module;

/* Holds info for the signal handler code, so it'll know how to shut down
 * jserv. */
static jserv_conf jserv_config;

void *create_jserv_config(pool * p, server_rec * s)
{
    jserv_conf *j = (jserv_conf *) pcalloc(p, sizeof(jserv_conf));
    time_t now;
    unsigned long bits1, bits2;

    j->dirs = make_array(p, 20, sizeof(jserv_entry));
    j->envs = make_array(p, 5, sizeof(char *));
    j->vhost = make_array(p, 10, sizeof(jserv_conf *));

    /* We don't store only the hostname because it may not have been set 
     * at this time
     */
    j->host = s;

    j->classpath = NULL;

    j->binary = NULL;
    j->binary_args = make_array(p, 10, sizeof(char *));

    j->properties = NULL;
    j->errlog = NULL;
    j->port = JSERV_DEFAULT_PORT;
    j->remote_server = JSERV_DEFAULT_SERVER;

    /* Create an auth string */
    now = time(NULL);

    /* use bits from moving targets */
    bits1 = (MODULE_MAGIC_NUMBER << 4);
    bits1 ^= getpid() << 17;	/* randomized by our PID */

    /* make a seed */
    srand(now | getppid() | bits1);	/* randomized by time and parent PID */

    /* and a randomized number */
    bits2 = ((unsigned long) rand()) * getpid() + rand();	/* overflow OK */
    bits2 ^= now;

    /* plus some obfuscation */
    /* :-) numbers chosen from the Java Bytecode magic numbers */
    bits1 ^= 0xCAFE;
    bits2 ^= 0xBABE;

    ap_snprintf(j->auth, sizeof(j->auth), "%04x%04x", bits1, bits2);

    strncpy(jserv_config.auth, j->auth, sizeof(jserv_config.auth));

    j->manual = 0;

    return j;
}

/** Merge per-virtual-host configuration information with the default
 * configuration for all hosts. 
 *
 * Returns: a pointer to a jserv_conf structure containing the 
 * results of the merge.
 **/
void *merge_jserv_config(pool * p, void *basev, void *overridesv)
{
    jserv_conf *j = (jserv_conf *) pcalloc(p, sizeof(jserv_conf));
    jserv_conf *base = (jserv_conf *) basev;
    jserv_conf *overrides = (jserv_conf *) overridesv;

    /* Put a pointer to the virtual host configuration in the base server */
    jserv_conf **new = (jserv_conf **) push_array(base->vhost);
    *new = j;

    /* In general, virtual directories can override the options
      * specified as the default. */
    j->dirs = overrides->dirs ? overrides->dirs : base->dirs;
    j->properties = (overrides->properties ? 
		     overrides->properties : base->properties );
    j->host = overrides->host ? overrides->host : base->host;
 
    /* However, Apache runs all servlets from all virtual hosts and
      * repositories in a single JVM, so it doesn't make sense to let
      * people override these properties in a VHost configuration.  */
    j->binary = base->binary;
    j->binary_args = base->binary_args;
    j->auth_export = base->auth_export;
    j->port = base->port;
    j->classpath = base->classpath;
    j->envs =  base->envs;
    j->errlog = base->errlog;

    strcpy(j->auth, base->auth);

    return j;
}

const char *add_jserv_dir(cmd_parms * cmd, void *dummy, char *f, char *r)
{
    server_rec *s = cmd->server;
    jserv_conf *conf =
    (jserv_conf *) get_module_config(s->module_config, &jserv_module);
    jserv_entry *new = push_array(conf->dirs);
    struct stat finfo;

    if (stat(r, &finfo) == -1)
	return pstrcat(cmd->pool, "\"", r, "\" does not exist", NULL);

    new->fake = f;
    new->real = r;
    new->is_dir = S_ISDIR(finfo.st_mode);

    /*  Not anymore */
    /* Add the dir to the classpath */
    /*     if (conf->classpath)
       conf->classpath = pstrcat(cmd->pool, r, CP_SEP, conf->classpath, NULL); 
       else 
       conf->classpath = r; */

    return NULL;
}

const char *set_jserv_classpath(cmd_parms * cmd, void *dummy, char *path)
{
    server_rec *s = cmd->server;
    jserv_conf *conf =
    (jserv_conf *) get_module_config(s->module_config, &jserv_module);

    if (conf->classpath)
	conf->classpath = pstrcat(cmd->pool, path, CP_SEP, conf->classpath,
				  NULL);
    else
	conf->classpath = path;

    return NULL;
}

const char *set_jserv_bin(cmd_parms * cmd, void *dummy, char *bin)
{
    server_rec *s = cmd->server;
    jserv_conf *conf =
    (jserv_conf *) get_module_config(s->module_config, &jserv_module);

    conf->binary = bin;

    return NULL;
}

const char *set_jserv_properties(cmd_parms * cmd, void *dummy, char *props)
{
    server_rec *s = cmd->server;
    jserv_conf *conf =
    (jserv_conf *) get_module_config(s->module_config, &jserv_module);

    conf->properties = server_root_relative(cmd->pool, props);

    return NULL;
}

const char *set_servlet_host (cmd_parms * cmd, void *dummy, char *host) {
    server_rec *s = cmd->server;
    jserv_conf *conf =
        (jserv_conf *)get_module_config(s->module_config,&jserv_module);

    struct hostent *hp;
    u_long ipaddr;
    int i;

    if (!host)
        return "you must provide a hostname in the ServletHost directive";

    for (i=0; host[i] != '\0'; i++)
        if (!isdigit(host[i]) && host[i] != '.')
            break;

    if (host[i] != '\0')
    {
        hp = gethostbyname(host);
        if (hp == NULL)
            return "Host not found";
        ipaddr = ((struct in_addr*)hp->h_addr_list[0])->s_addr;
    } else
    {
        ipaddr = inet_addr(host);
    }

    conf->remote_server = ipaddr;
    return NULL;
}

const char *set_jserv_errlog(cmd_parms * cmd, void *dummy, char *fname)
{
    server_rec *s = cmd->server;
    jserv_conf *conf =
    (jserv_conf *) get_module_config(s->module_config, &jserv_module);

    conf->errlog = fname;

    return NULL;
}

const char *set_jserv_env(cmd_parms * cmd, void *dummy, char *varname)
{
    server_rec *s = cmd->server;
    jserv_conf *conf =
    (jserv_conf *) get_module_config(s->module_config, &jserv_module);
    char **new = (char **) push_array(conf->envs);

    *new = varname;

    return NULL;
}

const char *set_jserv_arg(cmd_parms * cmd, void *dummy, char *arg)
{
    server_rec *s = cmd->server;
    jserv_conf *conf =
    (jserv_conf *) get_module_config(s->module_config, &jserv_module);
    char **new = (char **) push_array(conf->binary_args);

    *new = arg;

    return NULL;
}

const char *set_jserv_port(cmd_parms * cmd, void *dummy, char *port)
{
    server_rec *s = cmd->server;
    jserv_conf *conf =
    (jserv_conf *) get_module_config(s->module_config, &jserv_module);

    conf->port = atoi(port);

    return NULL;
}

const char *set_jserv_manual(cmd_parms * cmd, void *dummy, int arg)
{
    server_rec *s = cmd->server;
    jserv_conf *conf =
    (jserv_conf *) get_module_config(s->module_config, &jserv_module);

    conf->manual = arg;

    return NULL;
}

const char *set_auth_export(cmd_parms * cmd, void *dummy, int arg)
{
    server_rec *s = cmd->server;
    jserv_conf *conf =
    (jserv_conf *) get_module_config(s->module_config, &jserv_module);

    conf->auth_export = arg;

    return NULL;
}

command_rec jserv_cmds[] =
{
    {"ServletAlias", add_jserv_dir, NULL, RSRC_CONF, TAKE2,
     "a URL path prefix and a directory which stores class files"},
    {"ServletAuthExport", set_auth_export, NULL, RSRC_CONF, FLAG,
   "flag whether mod_jserv may let other modules access confidential info"},
    {"ServletBinary", set_jserv_bin, NULL, RSRC_CONF, TAKE1,
     "the path to the Java runtime binary"},
    {"ServletClassPath", set_jserv_classpath, NULL, RSRC_CONF, TAKE1,
     "a Java class path to search"},
    {"ServletEnvironment", set_jserv_env, NULL, RSRC_CONF, ITERATE,
     "A list of environment variables to pass to JServ"},
    {"ServletBinaryArgument", set_jserv_arg, NULL, RSRC_CONF, ITERATE,
     "A list of arguments to give to the Java interpreter."},
    {"ServletErrorLog", set_jserv_errlog, NULL, RSRC_CONF, TAKE1,
     "a filename to log System.err output from servlets to"},
    {"ServletManual", set_jserv_manual, NULL, RSRC_CONF, FLAG,
     "whether or not to run JServHandler manually"},
    {"ServletPort", set_jserv_port, NULL, RSRC_CONF, TAKE1,
     "the port to run the Java servlet server on"},
    {"ServletProperties", set_jserv_properties, NULL, RSRC_CONF, TAKE1,
     "the filename that holds the Java servlet properties"},
    {"ServletHost", set_servlet_host, NULL, RSRC_CONF, TAKE1,
     "the name of the host where JServ is running, if it's not local"},
    {NULL}
};

typedef struct {
    jserv_conf *conf;
    server_rec *s;
    pool *p;
    int log_fd;
} jserv_child_conf;

/* Child #2 - This is created by child #1 */

#if MODULE_MAGIC_NUMBER < 19970623
static void jserv_child_child(void *data)
{
#else
static int jserv_child_child(void *data)
{
#endif
    jserv_conf *conf = ((jserv_child_conf *) data)->conf;
    server_rec *s = ((jserv_child_conf *) data)->s;
    pool *p = ((jserv_child_conf *) data)->p;
    int i, log_fd = ((jserv_child_conf *) data)->log_fd;

    char port[10];
    char **env;
    char **args;


    if (log_fd >= 0) {
	args = malloc((sizeof(char *)) * (5 + conf->binary_args->nelts));
    }
    else {
	args = malloc((sizeof(char *)) * (4 + conf->binary_args->nelts));
    }

    /** 
     * Set command line option 
     * java [ java-option ] org.apache.jserv.JServHandler [ -t ] port
     */
    args[0] = conf->binary;

    /* Pass extra options to Java */
    for (i = 1; i < conf->binary_args->nelts + 1; i++) {
	char *arg = ((char **) (conf->binary_args->elts))[i - 1];

	args[i] = arg;
    }


    args[i] = JSERV_CLASS;
    i++;

    /* Add trace option if necessary */
    if (log_fd >= 0) {
	args[i] = "-t";
	i++;
    }

    /* Port */
    ap_snprintf(port, sizeof(port), "%d", conf->port);
    args[i] = pstrdup(p, port);
    i++;

    args[i] = NULL;

/** Set up the Environment */

    env = malloc((sizeof(char *)) * (4 + conf->envs->nelts));

    env[0] = pstrcat(p, "CLASSPATH=", conf->classpath, NULL);
    env[1] = pstrcat(p, "PATH=", getenv("PATH"), NULL);
    env[2] = pstrcat(p, "TZ=", getenv("TZ"), NULL);

    for (i = 0; i < conf->envs->nelts;) {
	char *var = ((char **) (conf->envs->elts))[i];

	env[3 + i] = pstrcat(p, var, "=", getenv(var), NULL);
	i++;
    }

    env[3 + i] = NULL;

    /* Change uid to the server's User, if we're running as root */
    if (!geteuid() && setuid(s->server_uid) == -1) {
	log_unixerr("setuid", NULL, "unable to change uid for JServ", s);
	exit(1);
    }

    if (log_fd >= 0) {
	dup2(log_fd, STDERR_FILENO);
    }
    else {
	error_log2stderr(s);
    }

    cleanup_for_exec();

    if (execve(conf->binary, args, env) < 0) {
	fprintf(stderr, "httpd: Jserv failed to exec Java interpreter.");
	exit(1);
    }

#if MODULE_MAGIC_NUMBER >= 19970623
    return 0;
#endif
}

/* Send a line in the format JServ expects; four hex characters indicating
 * length, then the string.
 */

static void send_hexline(FILE *f, const char *line) {
    fprintf(f, "%04x%s", strlen(line), line);
}

/* Used to send data to jserv if sockets fail */
static FILE *jserv_in = NULL;

static int count = 0;
static signal_current;
/* These is a signal handler for child #1. When it gets a signal,
 * it passes it on to JServ.
 */

static void jserv_signal(int sig) {
    sigset_t blocked_sigs;
    char sigstring[4];
    char authstring[JSERV_AUTH_CONF_SIZE + 2];
    FILE * j;
    int i, sock, use_stdin = 0;
    struct sockaddr_in addr;

    sigemptyset(&blocked_sigs);
    sigaddset(&blocked_sigs, SIGTERM);
    sigaddset(&blocked_sigs, SIGHUP);

    sigprocmask(SIG_BLOCK, &blocked_sigs, NULL);

    if (sig == SIGALRM)
        sig = signal_current;
    addr.sin_addr.s_addr = jserv_config.remote_server;
    addr.sin_port = htons(jserv_config.port);
    addr.sin_family = AF_INET;

    if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
        fprintf(stderr,"Error creating socket to terminate jserv process.\n");
        /* we're not going to be able to connect, so use the STDIN method*/
        use_stdin = 1;
    }
    else {
        do {
              i = connect(sock, (struct sockaddr *)&addr,
                                 sizeof(struct sockaddr_in));
#ifdef WIN32
              if(i == SOCKET_ERROR)
                  errno = WSAGetLastError() - WSABASEERR;
#endif /* WIN32 */
        }
        while (i == -1 && errno == EINTR);
        if (i == -1) {
            /* we weren't able to connect, so we use the STDIN method*/
            use_stdin = 1;
        }
    }

    if (use_stdin) {
        /* give it two tries first: */
        if (++count <= 2) {
	    signal_current = sig;
	    signal(SIGALRM, jserv_signal);
	    sleep(1);
	}
        fprintf(jserv_in, "S%02d", sig);
        fflush(jserv_in);
    }
    else {
        j = fdopen(sock, "rb+");

        authstring[0]='A';
        authstring[1]='\0';
        strncat (authstring, jserv_config.auth, JSERV_AUTH_CONF_SIZE);
        send_hexline(j, authstring);

        /* format up the message */
        sprintf(sigstring, "s%02d", sig);

        /* sent the signal encoded as a string */
        send_hexline(j, sigstring);

        /* Send the terminating entry */
        send_hexline(j, "");

        /* clean up */
        fflush(j);
        fclose(j);
    }

    if (sig == SIGTERM) {
      if (jserv_in)
          fclose(jserv_in);
      waitpid(-1, NULL, 0);
      exit(0);
    }
    count=0;
    sigprocmask(SIG_UNBLOCK, &blocked_sigs, NULL);
}

/* Send the configuration informations per host in the proper form 

 * hostname\tproperty_file(\tservlet_dir)+\n
 *
 */
static void send_host_config(pool * p, jserv_conf * conf)
{
    jserv_entry *entries = (jserv_entry *) conf->dirs->elts;
    int i;

    /* Verify that there is a servlet alias for that host
     */
    if (!(conf->dirs->nelts)) {
	log_error("mod_jserv: missing ServletAlias or ServletProperties directive.", conf->host);
	return;
    }

    /* Use default properties */
    if (!conf->properties) {
	conf->properties = server_root_relative(p, JSERV_DEFAULT_PROPERTIES);
    }

    fprintf(jserv_in, "%s\t%s", conf->host->server_hostname,
	    conf->properties);
    for (i = 0; i < conf->dirs->nelts; i++) {
	jserv_entry *dir = &entries[i];
	fprintf(jserv_in, "\t%s", dir->real);
    }

    fprintf(jserv_in, "\n");
    fflush(jserv_in);
}

/* Child #1 - This is created by Apache */

#if MODULE_MAGIC_NUMBER < 19970623
static void jserv_child(void *data)
{
#else
static int jserv_child(void *data)
{
#endif
    jserv_conf *conf = ((jserv_child_conf *) data)->conf;
    pool *p = ((jserv_child_conf *) data)->p;

    sigset_t blocked_sigs;
    int pid;
    int v;

    jserv_conf *host;

    /* Why do I do this? It's a long story. Apache does a killpg(SIGKILL)
     * when it exits via a TERM signal, so it kills all its children with
     * no chance for recourse. So I change process group to avoid that.
     *
     * Note that this is fixed in Apache 1.3b1
     */
#if MODULE_MAGIC_NUMBER < 19970902
    setpgid(0, 0);
#endif

    sigemptyset(&blocked_sigs);
    sigaddset(&blocked_sigs, SIGTERM);
    sigaddset(&blocked_sigs, SIGHUP);

    do {
	/* Block signals to prevent a race condition when starting
	 * child #2 (we recieve a SIGTERM after starting Java, but before
	 * we set up the signal handlers, so Java gets stuck in limbo
	 * forever).
	 */
	sigprocmask(SIG_BLOCK, &blocked_sigs, NULL);

	if (jserv_in) {
	    fclose(jserv_in);
	    free(jserv_in);
	}

	/* Spawn child #2 */
	if (!(pid = spawn_child(p, jserv_child_child, data, kill_never,
				&jserv_in, NULL))) {
	    fprintf(stderr, "httpd: could not spawn JServ child process");
	    exit(1);
	}

	/* Send the auth data to JServ */
	fprintf(jserv_in, "%s\n", conf->auth);
	fflush(jserv_in);

	/* Send base configuration */
	send_host_config(p, conf);

	/* Send virtual configuration */
	for (v = 0; v < conf->vhost->nelts; v++) {
	    host = ((jserv_conf **) conf->vhost->elts)[v];
	    send_host_config(p, host);
	}

	/* Terminate the sending of data */
	fprintf(jserv_in, "end\n");
	fflush(jserv_in);

	/* Set up signal handlers */
	signal(SIGTERM, jserv_signal);
	signal(SIGHUP, jserv_signal);

	/* Unblock the signals, so we can receive them */
	sigprocmask(SIG_UNBLOCK, &blocked_sigs, NULL);

	/* Poll for a change. Every second, check the following items
	 * to see if we should do something:
	 */
	for (;;) {
	    /* Wait for a sec */
	    sleep(1);

	    /* Check our parent. If it's 1, it means we're alone,
	     * and should die.
	     */
	    if (getppid() == 1)
		jserv_signal(SIGTERM);

	    /* Check on the child pid. If the status changed, it means
	     * that JServ has exited, and we should reinstall it.
	     */
	    if (waitpid(pid, NULL, WNOHANG) == pid)
		break;
	}

	/* Sleep for a few secs before reinstalling Java */
	sleep(3);
    } while (1);
}

/** Initialize the module: called once when Apache starts.
 *
 * After initializing data, we fork a child which will in turn
 * fork and monitor a JVM.
 **/
void init_jserv(server_rec * s, pool * p)
{
    jserv_conf *conf =
    (jserv_conf *) get_module_config(s->module_config, &jserv_module);
    jserv_child_conf *jcc;

    /* Are we correctly configured for automatic startup? */
    if (!conf->binary || !conf->classpath || conf->manual)
	return;

    /* Property file location are passed by sdtin */
    /*if (!conf->properties)
       conf->properties = server_root_relative(p, JSERV_DEFAULT_PROPERTIES);
     */

    if (conf->remote_server != JSERV_DEFAULT_SERVER) {
	fprintf(stderr, "http: you cannot use ServletHost in automatic mode; use 'ServletManual on' (and set up a server manually) to use manual mode\n");
	exit(1);
    }

    /* Setup child data */
    jcc = pcalloc(p, sizeof(jserv_child_conf));
    jcc->conf = conf;
    jcc->s = s;
    jcc->p = p;

    /* Open error log */
    if (conf->errlog) {
	char *fname = server_root_relative(p, conf->errlog);
	if ((jcc->log_fd = popenf(p, fname, xfer_flags, xfer_mode)) < 0) {
	    perror("open");
	    fprintf(stderr, "http: could not open JServ error log file\n");
	    exit(1);
	}
    }
    else {
	jcc->log_fd = -1;
    }

    jserv_config.port = -1;
    jserv_config.remote_server = (unsigned long) -1;
    memcpy(&jserv_config, conf, sizeof(jserv_conf));

    /* Spawn child */
    if (!(spawn_child(p, jserv_child, (void *) jcc, kill_after_timeout,
		      NULL, NULL))) {
	fprintf(stderr, "httpd: could not spawn JServ child process");
	exit(1);
    }
}

static int jserv_alias_matches(char *uri, char *alias_fakename)
{
    char *end_fakename = alias_fakename + strlen(alias_fakename);
    char *aliasp = alias_fakename, *urip = uri;

    while (aliasp < end_fakename) {
	if (*aliasp == '/') {
	    /* any number of '/' in the alias matches any number in
	     * the supplied URI, but there must be at least one...
	     */
	    if (*urip != '/')
		return 0;

	    while (*aliasp == '/')
		++aliasp;
	    while (*urip == '/')
		++urip;
	}
	else {
	    /* Other characters are compared literally */
	    if (*urip++ != *aliasp++)
		return 0;
	}
    }

    /* Check last alias path component matched all the way */

    if (aliasp[-1] != '/' && *urip != '\0' && *urip != '/')
	return 0;

    /* Return number of characters from URI which matched (may be
     * greater than length of alias, since we may have matched
     * doubled slashes)
     */

    return urip - uri;
}

static char *try_jserv_list(request_rec * r, array_header * dirs)
{
    jserv_entry *entries = (jserv_entry *) dirs->elts;
    int i;

    for (i = 0; i < dirs->nelts; ++i) {
	jserv_entry *e = &entries[i];
	int l = jserv_alias_matches(r->uri, e->fake);

	if (l > 0) {
	    /* Note some info for later use */
	    table_set(r->notes, "jserv-dirname", e->real);
	    if (e->is_dir)
		table_set(r->notes, "jserv-isdir", "true");
	    r->handler = pstrdup(r->pool, "jserv-servlet");

	    return pstrcat(r->pool, e->real, r->uri + l, NULL);
	}
    }

    return NULL;
}

int translate_jserv(request_rec * r)
{
    void *sc = r->server->module_config;
    jserv_conf *conf = (jserv_conf *) get_module_config(sc, &jserv_module);
    char *ret;

#if defined(__EMX__) || defined(WIN32)
    /* Add support for OS/2 drive names */
    if ((r->uri[0] != '/' && r->uri[0] != '\0') && r->uri[1] != ':')
#else
    if (r->uri[0] != '/' && r->uri[0] != '\0')
#endif
	return DECLINED;

    if ((ret = try_jserv_list(r, conf->dirs))) {
	r->filename = ret;
	return OK;
    }
    return DECLINED;
}

/* export info early, during the header parsing stage */
int jserv_req_init(request_rec * r)
{
    void *sc = r->server->module_config;
    jserv_conf *conf = (jserv_conf *) get_module_config(sc, &jserv_module);

    /* some conditional notes for other modules (i.e. perl, php) */
    if (conf->auth_export) {
	char port[15];

	ap_snprintf(port, sizeof(port), "%d", conf->port);
	table_set(r->notes, "jserv-port", pstrdup(r->pool, port));
	table_set(r->notes, "jserv-auth", pstrdup(r->pool, conf->auth));
    }
    return DECLINED;
}

int jserv_handler(request_rec * r)
{
    void *s = r->server->module_config;
    jserv_conf *conf = (jserv_conf *) get_module_config(s, &jserv_module);
    char *p, *classname, *dirname = table_get(r->notes, "jserv-dirname");
    char *hdr, *is_dir = table_get(r->notes, "jserv-isdir");
    int ret;

    struct sockaddr_in addr;
    FILE *j;
    int i, sock;

    /* Make sure we're set up correctly */
    if (!conf->binary && !conf->manual ) {
	log_reason("ServletBinary not configured", r->uri, r);
	return SERVER_ERROR;
    }
    if (!conf->classpath && !conf->manual) {
	log_reason("ServletClassPath not configured", r->uri, r);
	return SERVER_ERROR;
    }

    /* Make sure we have a dir name - we must be called with
     * a matching ServletAlias directive, found when translating.
     */
    if (!dirname) {
	log_reason("Servlet class name unknown (illegal access?)", r->uri, r);
	return FORBIDDEN;
    }

    /* If it's a directory, then this is simple. If not, this is more
     * complex, since we need to identify the class name from the path info
     */

    if (is_dir) {
	classname = pstrdup(r->pool, r->filename + strlen(dirname) + 1);

	/* Translate slash (/) to period (.), so you can use physical
	 * names to identify classes.
	 */
	for (p = classname; *p; p++) {
	    if (*p == '/')
		*p = '.';
	}
    }
    else if (!r->path_info) {
	/* We should have some path info... if not, it means we didn't
	 * get a class to match!
	 */
	return NOT_FOUND;
    }
    else {
	char *path = r->path_info;
	char *slash;

	/* If there's a leading /, strip it off */
	if (path[0] == '/')
	    path++;

	slash = strchr(path, '/');

	if (!slash) {
	    /* No extra path info */
	    classname = pstrdup(r->pool, path);
	    r->path_info = "";
	}
	else {
	    classname = pstrndup(r->pool, path, slash - path);
	    r->path_info = slash;
	}
    }

    /* Make sure we have a class name. */

    if (!classname || !*classname) {
	log_reason("Class name could not be identified", r->uri, r);
	return NOT_FOUND;
    }

    /* Open the connection to the server */

    addr.sin_addr.s_addr = conf->remote_server;
    addr.sin_port = htons(conf->port);
    addr.sin_family = AF_INET;

    if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
	log_reason("Could not create JServ socket", r->uri, r);
	return SERVER_ERROR;
    }
    note_cleanups_for_fd(r->pool, sock);

    hard_timeout("proxy connect", r);

    do {
	i = connect(sock, (struct sockaddr *) &addr,
		    sizeof(struct sockaddr_in));
#ifdef WIN32
	if (i == SOCKET_ERROR)
	    errno = WSAGetLastError() - WSABASEERR;
#endif /* WIN32 */
    }
    while (i == -1 && errno == EINTR);
    if (i == -1) {
	log_reason("JServ connect failed", r->uri, r);
	return SERVER_ERROR;
    }

    kill_timeout(r);

    j = pfdopen(r->pool, sock, "rb+");

    /* Send the data */

    add_cgi_vars(r);
    add_common_vars(r);

    if ((ret = setup_client_block(r, REQUEST_CHUNKED_ERROR)))
	return ret;

    hard_timeout("jserv send", r);

    /* We don't yet have auth code, so send "auth" */
    send_hexline(j, pstrcat(r->pool, "A", conf->auth, NULL));

    /* Send servlet alias matched and class name : C = class */
    send_hexline(j, pstrcat(r->pool, "C", dirname, "\t", classname, NULL));

    /* Send host name S = Server */
    send_hexline(j, pstrcat(r->pool, "S", conf->host->server_hostname, NULL));

    /* Send environment vars, except for the header ones */
    if (r->subprocess_env) {
	array_header *env_arr = table_elts(r->subprocess_env);
	table_entry *elts = (table_entry *) env_arr->elts;

	for (i = 0; i < env_arr->nelts; ++i) {
	    if (!elts[i].key)
		continue;
	    if (!strncmp(elts[i].key, "HTTP_", 5))
		continue;
	    send_hexline(j, pstrcat(r->pool, "E", elts[i].key, "\t",
				    elts[i].val, NULL));
	}
    }

    /* Send the request headers */

    if (r->headers_in) {
	array_header *hdr_arr = table_elts(r->headers_in);
	table_entry *elts = (table_entry *) hdr_arr->elts;

	for (i = 0; i < hdr_arr->nelts; ++i) {
	    if (!elts[i].key)
		continue;
	    send_hexline(j, pstrcat(r->pool, "H", elts[i].key, "\t",
				    elts[i].val, NULL));
	}
    }

    /* Send the terminating entry */
    send_hexline(j, "");

    /* If there is a request entity, send it */

    if (should_client_block(r)) {
	size_t len_read;
	char buffer[HUGE_STRING_LEN];

	while ((len_read = get_client_block(r, buffer, HUGE_STRING_LEN)) > 0) {
	    reset_timeout(r);
	    if (fwrite(buffer, sizeof(char), len_read, j) < len_read) {
		/* Something happened, soak up the rest */
		while (get_client_block(r, buffer, HUGE_STRING_LEN) > 0);
		break;
	    }
	}
    }

    fflush(j);
    kill_timeout(r);

    /* Receive the response */

    hard_timeout("jserv receive", r);

    if ((ret = scan_script_header(r, j))) {
	kill_timeout(r);
	return ret;
    }

    /* Check for our special headers */

    if ((hdr = table_get(r->err_headers_out, "Servlet-Log"))) {
	table_unset(r->err_headers_out, "Servlet-Log");
	log_error(hdr, r->server);
    }

    if ((hdr = table_get(r->err_headers_out, "Servlet-Error"))) {
	int status = r->status;

	table_unset(r->err_headers_out, "Servlet-Error");
	log_reason(hdr, r->uri, r);

	/* Zap the existing status, or Apache will do odd things */
	r->status = HTTP_OK;
	r->status_line = NULL;

	return status;
    }

    /* Do CGI-style Location-handling */

    if (table_get(r->headers_out, "Location") && r->status == 200) {
	kill_timeout(r);

	r->status = HTTP_OK;
	r->status_line = NULL;

	return REDIRECT;
    }

    send_http_header(r);
    if (!r->header_only)
	send_fd(j, r);

    kill_timeout(r);
    pfclose(r->pool, j);

    return OK;
}

handler_rec jserv_handlers[] =
{
    {"jserv-servlet", jserv_handler},
    {NULL}
};

module jserv_module =
{
    STANDARD_MODULE_STUFF,
    init_jserv,			/* initializer */
    NULL,			/* dir config creater */
    NULL,			/* dir merger --- default is to override */
    create_jserv_config,	/* server config */
    merge_jserv_config,		/* merge server configs */
    jserv_cmds,			/* command table */
    jserv_handlers,		/* handlers */
    translate_jserv,		/* filename translation */
    NULL,			/* check_user_id */
    NULL,			/* check auth */
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
    NULL,			/* logger */
    jserv_req_init		/* header parser */
};
